# -*- coding: utf-8; -*-

# This file is part of python-ads1115.
#
# python-ads1115 is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# python-ads1115 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with python-ads1115.  If not, see <http://www.gnu.org/licenses/>

# 2018-01-17 David Goncalves <david@lestat.st>
#            Python class to communicate with an ADS1115 chip on I2C bus.
#            It's a Texas Instruments 4 ports 16bit ADC.

import smbus
import sys
import time


class ads1115:

    # Internal registers
    REG_CONVERT = 0x00
    REG_CONFIG = 0x01
    REG_LOW_THRESH = 0x02
    REG_HIGH_THRESH = 0x03

    # Configuration register parameters
    # OS = Operational Status
    CONF_OS_SINGLE = 0x1 << 15           # Write: Start a single conversion (when in power-down state)
    CONF_OS_BUSY = 0x1 << 15             # Read: Device is currently doing a conversion
    CONF_OS_IDLE = 0x0                   # Read: Device is idle

    # MUX = Input Multiplexer
    CONF_MUX_DIFF_AIN0_AIN1 = 0x0        # Input differential P=AIN0 N=AIN1 (default)
    CONF_MUX_DIFF_AIN0_AIN3 = 0x1 << 12  # Input differential P=AIN0 N=AIN3
    CONF_MUX_DIFF_AIN1_AIN3 = 0x2 << 12  # Input differential P=AIN1 N=AIN3
    CONF_MUX_DIFF_AIN2_AIN3 = 0x3 << 12  # Input differential P=AIN2 N=AIN3
    CONF_MUX_SINGLE_AIN0 = 0x4 << 12     # Input single P=AIN0 N=GND
    CONF_MUX_SINGLE_AIN1 = 0x5 << 12     # Input single P=AIN1 N=GND
    CONF_MUX_SINGLE_AIN2 = 0x6 << 12     # Input single P=AIN2 N=GND
    CONF_MUX_SINGLE_AIN3 = 0x7 << 12     # Input single P=AIN3 N=GND

    diff_map = {'0-1': CONF_MUX_DIFF_AIN0_AIN1,
                '0-3': CONF_MUX_DIFF_AIN0_AIN3,
                '1-3': CONF_MUX_DIFF_AIN1_AIN3,
                '2-3': CONF_MUX_DIFF_AIN2_AIN3}

    # PGA = Programmable Gain Amplifier (to determine the Full Scale Range)
    CONF_PGA_6144MV = 0x0                # FSR = 6144mV
    CONF_PGA_4096MV = 0x1 << 9           # FSR = 4096mV
    CONF_PGA_2048MV = 0x2 << 9           # FSR = 2048mV (default)
    CONF_PGA_1024MV = 0x3 << 9           # FSR = 1024mV
    CONF_PGA_512MV = 0x4 << 9            # FSR = 512mV
    CONF_PGA_256MV = 0x5 << 9            # FSR = 256mV

    fsr_map = {6144: CONF_PGA_6144MV,
               4096: CONF_PGA_4096MV,
               2048: CONF_PGA_2048MV,
               1024: CONF_PGA_1024MV,
               512: CONF_PGA_512MV,
               256: CONF_PGA_512MV}

    # MODE = Device operating mode
    CONF_MODE_CONTINUOUS = 0x0           # Continuous conversion mode
    CONF_MODE_SINGLE = 0x1 << 8          # Single shot mode (default)

    # DR = Data Rate (samples per second)
    CONF_DR_8SPS = 0x0                   # 8 samples/s
    CONF_DR_16SPS = 0x1 << 5             # 16 samples/s
    CONF_DR_32SPS = 0x2 << 5             # 32 samples/s
    CONF_DR_64SPS = 0x3 << 5             # 64 samples/s
    CONF_DR_128SPS = 0x4 << 5            # 128 samples/s (default)
    CONF_DR_250SPS = 0x5 << 5            # 250 samples/s
    CONF_DR_475SPS = 0x6 << 5            # 475 samples/s
    CONF_DR_860SPS = 0x7 << 5            # 860 samples/s

    sps_map = {8: CONF_DR_8SPS,
               16: CONF_DR_16SPS,
               32: CONF_DR_32SPS,
               64: CONF_DR_64SPS,
               128: CONF_DR_128SPS,
               250: CONF_DR_250SPS,
               475: CONF_DR_475SPS,
               860: CONF_DR_860SPS}

    # COMPMODE = Comparator mode
    CONF_COMPMODE_TRADITIONAL = 0x0      # Traditional comparator (default)
    CONF_COMPMODE_WINDOW = 0x1 << 4      # Window comparator

    # COMPPOL = Comparator polarity
    CONF_COMPPOL_LOW = 0x0               # Active low (default)
    CONF_COMPPOL_HIGH = 0x1 << 3         # Active high

    # COMPLAT = Comparator latching (control the ALERT/RDY pin)
    CONF_COMPLAT_DISABLED = 0x0          # ALERT/RDY pin does not latch when asserted (default)
    CONF_COMPLAT_ENABLED = 0x1 << 2      # ALERT/RDY pin stay latched until data read by the master

    # COMPQUE = Comparator queue
    CONF_COMPQUE_1CONV = 0x0             # Assert after 1 conversion
    CONF_COMPQUE_2CONV = 0x1             # Assert after 2 conversions
    CONF_COMPQUE_4CONV = 0x2             # Assert after 4 conversions
    CONF_COMPQUE_DISABLED = 0x3          # Disable comparator and set ALERT/RDY to high impedence (default)

    def __init__(self, address=0x48, bus=1, debug=False):
        self._debug = debug
        self.i2c_addr = address
        self.i2c_bus = smbus.SMBus(bus)

        self._debug_message("Accessing ADS1115 at address 0x%x on bus %s" % (address, bus))

    def _debug_message(self, message="", error=False):
        """
        Prints debug messages to console when debuging mode is activated
        """

        if (self._debug is True) or (error is True):
            print("%s %s" % (("[DEBUG]" if error is False else "[ERROR]"), message))
            sys.stdout.flush()
            if (error is True):
                raise Exception(message)

    def _write_data(self, address, data):
        """
        Private method to write data to the ADS1115
        """

        self._debug_message("> Writing %s bytes to register 0x%x [%s]" % (len(data), address, ' '.join(["0x%02X" % x for x in data])))
        self.i2c_bus.write_i2c_block_data(self.i2c_addr, address, data)

    def _read_data(self, address, size):
        """
        Private method to read data from the ADS1115
        """

        self._debug_message("> Reading %s bytes from register 0x%x" % (size, address))
        data = self.i2c_bus.read_i2c_block_data(self.i2c_addr, address, size)
        self._debug_message("> ADS1115 response : [%s]" % ' '.join(["0x%02X" % x for x in data]))
        return(data)

    def ReadSingleEnded(self, port=0, fsr=6144, sps=860):
        """
        Start a new single ended conversion on chip and get the result (in mV)
        """

        if port not in range(0, 4):
            raise ValueError("Port must be in range 0 to 3")

        if fsr not in self.fsr_map.keys():
            raise ValueError("Wrong FSR '%s'. Valid values are %s" % (fsr, repr(self.fsr_map.keys())))

        if sps not in self.sps_map.keys():
            raise ValueError("Wrong SPS '%s'. Valid values are %s" % (sps, repr(self.sps_map.keys())))

        self._debug_message("Reading single ended measurement on port %s with FSR=%smV" % (port, fsr))

        # Configure the chip
        config = self.CONF_COMPQUE_DISABLED | \
            self.CONF_COMPLAT_DISABLED | \
            self.CONF_COMPPOL_LOW | \
            self.CONF_COMPMODE_TRADITIONAL | \
            self.CONF_MODE_SINGLE | \
            self.sps_map.get(sps) | \
            self.fsr_map.get(fsr) | \
            self.CONF_MUX_SINGLE_AIN0 + (port << 12) | \
            self.CONF_OS_SINGLE

        self._write_data(self.REG_CONFIG, [config >> 8, config & 0xFF])

        # Wait until conversion finishes... (we wait for 2 samples for safety)
        time.sleep((1.0 / sps) * 2)

        result = self._read_data(self.REG_CONVERT, 2)
        raw = result[0] << 8 | result[1]

        if (raw >> 15) == 1:
            raw -= 0xFFFF

        millivolt = raw * (fsr / 32768.0)
        self._debug_message("Voltage on port %s is %smV" % (port, millivolt))
        return(millivolt)

    def ReadDifferential(self, ports='0-1', fsr=6144, sps=860):
        """
        Start a new differential conversion on chip and get the result (in mV)
        """

        if ports not in self.diff_map.keys():
            raise ValueError("Wrong ports '%s'. Valid values are %s" % (fsr, repr(self.diff_map.keys())))

        if fsr not in self.fsr_map.keys():
            raise ValueError("Wrong FSR '%s'. Valid values are %s" % (fsr, repr(self.fsr_map.keys())))

        if sps not in self.sps_map.keys():
            raise ValueError("Wrong SPS '%s'. Valid values are %s" % (sps, repr(self.sps_map.keys())))

        self._debug_message("Reading differential measurement on ports %s with FSR=%smV" % (ports, fsr))

        # Configure the chip
        config = self.CONF_COMPQUE_DISABLED | \
            self.CONF_COMPLAT_DISABLED | \
            self.CONF_COMPPOL_LOW | \
            self.CONF_COMPMODE_TRADITIONAL | \
            self.CONF_MODE_SINGLE | \
            self.sps_map.get(sps) | \
            self.fsr_map.get(fsr) | \
            self.diff_map.get(ports) | \
            self.CONF_OS_SINGLE

        self._write_data(self.REG_CONFIG, [config >> 8, config & 0xFF])

        # Wait until conversion finishes... (we wait for 2 samples for safety)
        time.sleep((1.0 / sps) * 2)

        result = self._read_data(self.REG_CONVERT, 2)
        raw = result[0] << 8 | result[1]

        if (raw >> 15) == 1:
            raw -= 0xFFFF

        millivolt = raw * (fsr / 32768.0)
        self._debug_message("Differential voltage on ports %s is %smV" % (ports, millivolt))
        return(millivolt)
