# -*- coding: utf-8; -*-

# This file is part of python-ads1115.
#
# python-ads1115 is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# python-ads1115 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with python-ads1115.  If not, see <http://www.gnu.org/licenses/>

# 2018-01-17 David Goncalves <david@lestat.st>
#            Python class to communicate with an ADS1115 chip on I2C bus.
#            It's a Texas Instruments 4 ports 16bit ADC.
#
#            > Code ported to micropython
#            > This version has no debug/comments to provide a smaller module

import sys
import time
import ustruct
from micropython import const


class uads1115:

    REG_CONVERT = const(0x00)
    REG_CONFIG = const(0x01)
    REG_LOW_THRESH = const(0x02)
    REG_HIGH_THRESH = const(0x03)

    CONF_OS_SINGLE = const(0x1 << 15)
    CONF_OS_BUSY = const(0x1 << 15)
    CONF_OS_IDLE = const(0x0)

    CONF_MUX_DIFF_AIN0_AIN1 = const(0x0)
    CONF_MUX_DIFF_AIN0_AIN3 = const(0x1 << 12)
    CONF_MUX_DIFF_AIN1_AIN3 = const(0x2 << 12)
    CONF_MUX_DIFF_AIN2_AIN3 = const(0x3 << 12)
    CONF_MUX_SINGLE_AIN0 = const(0x4 << 12)
    CONF_MUX_SINGLE_AIN1 = const(0x5 << 12)
    CONF_MUX_SINGLE_AIN2 = const(0x6 << 12)
    CONF_MUX_SINGLE_AIN3 = const(0x7 << 12)

    diff_map = {'0-1': CONF_MUX_DIFF_AIN0_AIN1,
                '0-3': CONF_MUX_DIFF_AIN0_AIN3,
                '1-3': CONF_MUX_DIFF_AIN1_AIN3,
                '2-3': CONF_MUX_DIFF_AIN2_AIN3}

    CONF_PGA_6144MV = const(0x0)
    CONF_PGA_4096MV = const(0x1 << 9)
    CONF_PGA_2048MV = const(0x2 << 9)
    CONF_PGA_1024MV = const(0x3 << 9)
    CONF_PGA_512MV = const(0x4 << 9)
    CONF_PGA_256MV = const(0x5 << 9)

    fsr_map = {6144: CONF_PGA_6144MV,
               4096: CONF_PGA_4096MV,
               2048: CONF_PGA_2048MV,
               1024: CONF_PGA_1024MV,
               512: CONF_PGA_512MV,
               256: CONF_PGA_512MV}

    CONF_MODE_CONTINUOUS = const(0x0)
    CONF_MODE_SINGLE = const(0x1 << 8)

    CONF_DR_8SPS = const(0x0)
    CONF_DR_16SPS = const(0x1 << 5)
    CONF_DR_32SPS = const(0x2 << 5)
    CONF_DR_64SPS = const(0x3 << 5)
    CONF_DR_128SPS = const(0x4 << 5)
    CONF_DR_250SPS = const(0x5 << 5)
    CONF_DR_475SPS = const(0x6 << 5)
    CONF_DR_860SPS = const(0x7 << 5)

    sps_map = {8: CONF_DR_8SPS,
               16: CONF_DR_16SPS,
               32: CONF_DR_32SPS,
               64: CONF_DR_64SPS,
               128: CONF_DR_128SPS,
               250: CONF_DR_250SPS,
               475: CONF_DR_475SPS,
               860: CONF_DR_860SPS}

    CONF_COMPMODE_TRADITIONAL = const(0x0)
    CONF_COMPMODE_WINDOW = const(0x1 << 4)

    CONF_COMPPOL_LOW = const(0x0)
    CONF_COMPPOL_HIGH = const(0x1 << 3)

    CONF_COMPLAT_DISABLED = const(0x0)
    CONF_COMPLAT_ENABLED = const(0x1 << 2)

    CONF_COMPQUE_1CONV = const(0x0)
    CONF_COMPQUE_2CONV = const(0x1)
    CONF_COMPQUE_4CONV = const(0x2)
    CONF_COMPQUE_DISABLED = const(0x3)

    def __init__(self, bus, address=0x48):
        self.i2c_addr = address
        self.i2c_bus = bus

    def _write_data(self, address, data):
        self.i2c_bus.writeto_mem(self.i2c_addr, address, bytearray(data))

    def _read_data(self, address, size):
        return(self.i2c_bus.readfrom_mem(self.i2c_addr, address, size))

    def ReadSingleEnded(self, port=0, fsr=6144, sps=860):
        if port not in range(0, 4):
            raise ValueError("Port must be in range 0 to 3")

        if fsr not in self.fsr_map.keys():
            raise ValueError("Wrong FSR '%s'. Valid values are %s" % (fsr, repr(self.fsr_map.keys())))

        if sps not in self.sps_map.keys():
            raise ValueError("Wrong SPS '%s'. Valid values are %s" % (sps, repr(self.sps_map.keys())))

        config = self.CONF_COMPQUE_DISABLED | \
            self.CONF_COMPLAT_DISABLED | \
            self.CONF_COMPPOL_LOW | \
            self.CONF_COMPMODE_TRADITIONAL | \
            self.CONF_MODE_SINGLE | \
            self.sps_map.get(sps) | \
            self.fsr_map.get(fsr) | \
            self.CONF_MUX_SINGLE_AIN0 + (port << 12) | \
            self.CONF_OS_SINGLE

        self._write_data(self.REG_CONFIG, ustruct.pack('>H', config))
        time.sleep((1.0 / sps) * 2)

        raw = ustruct.unpack('>H', self._read_data(self.REG_CONVERT, 2))[0]

        if (raw >> 15) == 1:
            raw -= 0xFFFF

        return(raw * (fsr / 32768.0))

    def ReadDifferential(self, ports='0-1', fsr=6144, sps=860):
        if ports not in self.diff_map.keys():
            raise ValueError("Wrong ports '%s'. Valid values are %s" % (fsr, repr(self.diff_map.keys())))

        if fsr not in self.fsr_map.keys():
            raise ValueError("Wrong FSR '%s'. Valid values are %s" % (fsr, repr(self.fsr_map.keys())))

        if sps not in self.sps_map.keys():
            raise ValueError("Wrong SPS '%s'. Valid values are %s" % (sps, repr(self.sps_map.keys())))

        config = self.CONF_COMPQUE_DISABLED | \
            self.CONF_COMPLAT_DISABLED | \
            self.CONF_COMPPOL_LOW | \
            self.CONF_COMPMODE_TRADITIONAL | \
            self.CONF_MODE_SINGLE | \
            self.sps_map.get(sps) | \
            self.fsr_map.get(fsr) | \
            self.diff_map.get(ports) | \
            self.CONF_OS_SINGLE

        self._write_data(self.REG_CONFIG, ustruct.pack('>H', config))
        time.sleep((1.0 / sps) * 2)

        raw = ustruct.unpack('>H', self._read_data(self.REG_CONVERT, 2))[0]

        if (raw >> 15) == 1:
            raw -= 0xFFFF

        return(raw * (fsr / 32768.0))
